/*
Continuous Testing Toolkit
Copyright (C) 2009 Paul R. Holser, Jr.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

package org.infinitest.toolkit;

import static java.lang.Math.*;

import org.junit.Before;
import org.junit.Test;
import static org.hamcrest.Matchers.*;
import static org.junit.Assert.*;

/**
 * Support structure for testing a class's compliance with the {@link Comparable} interface.
 * <p/>
 * {@link #less()}, {@link #equal()}, and {@link #greater()} should return objects that are of the same class.
 *
 * @param <T> the type of the comparable objects
 * @author <a href="http://www.pholser.com">Paul Holser</a>
 */
public abstract class ComparabilityTestSupport<T extends Comparable<T>> {
    private T less;
    private T equal1;
    private T equal2;
    private T greater;

    /**
     * @return a new instance of the class under test; each object returned from this method should compare less than
     *         the objects returned from {@link #equal()}
     */
    protected abstract T less() throws Exception;

    /**
     * @return a new instance of the class under test; each object returned from this method should compare equal to
     *         each other
     */
    protected abstract T equal() throws Exception;

    /**
     * @return a new instance of the class under test; each object returned from this method should compare greater than
     *         the objects returned from {@link #equal()}
     */
    protected abstract T greater() throws Exception;

    @Before
    public final void initializeHarness() throws Exception {
        less = less();
        equal1 = equal();
        equal2 = equal();
        greater = greater();

        assertNotNull("createLessInstance() returned null", less);
        assertNotNull("createEqualInstance() returned null", equal1);
        assertNotNull("2nd createEqualInstance() returned null", equal2);
        assertNotNull("createGreaterInstance() returned null", greater);

        assertEquals("less and equal1 of different classes", less.getClass(), equal1.getClass());
        assertEquals("less and equal2 of different classes", less.getClass(), equal2.getClass());
        assertEquals("less and greater of different classes", less.getClass(), greater.getClass());

        checkForEquality(equal1, equal2);
    }

    /**
     * Override as a no-op if you do not require that {@link #equal()} return distinct but equivalent objects.
     */
    protected void checkForEquality(T first, T second) {
        assertNotSame("1st equal instance same as 2nd", first, second);
        assertEquals("1st equal not equal to 2nd", first, second);
    }

    @Test
    public final void reverseOfComparisonShouldGiveResultOfOppositeSign() {
        assertEquals("less vs. equal1", signum(less.compareTo(equal1)), -signum(equal1.compareTo(less)), 0);
        assertEquals("less vs. equal2", signum(less.compareTo(equal2)), -signum(equal2.compareTo(less)), 0);
        assertEquals("less vs. greater", signum(less.compareTo(greater)), -signum(greater.compareTo(less)), 0);
        assertEquals("equal1 vs. equal2", signum(equal1.compareTo(equal2)), -signum(equal2.compareTo(equal1)), 0);
        assertEquals("equal1 vs. greater", signum(equal1.compareTo(greater)), -signum(greater.compareTo(equal1)), 0);
        assertEquals("equal2 vs. greater", signum(equal2.compareTo(greater)), -signum(greater.compareTo(equal2)), 0);
    }

    @Test
    public final void twoObjectsWhichCompareLikeToAnotherObjectShouldGiveResultOfSameSign() {
        assertEquals("equal vs. less", signum(equal1.compareTo(less)), signum(equal2.compareTo(less)), 0);
        assertEquals("equal vs. greater", signum(equal1.compareTo(greater)), signum(equal2.compareTo(greater)), 0);
    }

    @Test
    public final void comparisonShouldGiveCorrectRelativeValue() {
        assertThat(less, lessThan(equal1));
        assertThat(less, lessThan(equal2));
        assertThat(greater, greaterThan(less));
        assertEquals(0, equal1.compareTo(equal2));
        assertThat(greater, greaterThan(equal1));
        assertThat(greater, greaterThan(equal2));
    }
}
