/*
Continuous Testing Toolkit
Copyright (C) 2009 Paul R. Holser, Jr.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

package org.infinitest.toolkit;

import java.lang.reflect.Field;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import static java.lang.reflect.Array.*;
import static java.util.Collections.*;

import org.junit.Before;
import org.junit.Test;
import static org.junit.Assert.*;

/**
 * Support structure for testing the {@linkplain Cloneable cloneability} of instances of a class.
 *
 * The contract for {@link Object#clone()} is fairly lax, but this class tests that after a clone operation, the
 * original and the clone references refer to instances of the same class, and are not the same, and compare
 * {@linkplain Object#equals(Object) equal}, and whose respective reference fields do not refer to the same
 * objects.  For reference fields that are arrays, {@link List}s, or {@link Map}s, tests also that each element
 * of the original collection does not refer to the same object as the corresponding slot in the clone (values
 * only for maps, since keys of maps should be immutable).
 *
 * @author <a href="http://www.pholser.com">Paul Holser</a>
 */
public abstract class StrictCloneabilityTestSupport<T extends Cloneable> {
    private T original;
    private T copy;
    private final Set<Class<?>> immutableClasses = new HashSet<Class<?>>();

    @Before
    public final void initializeHarness() throws Exception {
        original = instance();
        assertNotNull("instance() returned null", original);

        copy = copy(original);
        immutableClasses.addAll(coreImmutableClasses());
        immutableClasses.addAll(classesToTreatAsImmutable());
    }

    @Test
    public final void copyShouldBeOfSameClassAsOriginal() {
        assertEquals(original.getClass(), copy.getClass());
    }

    @Test
    public final void copyShouldNotBeIdenticalToOriginal() {
        assertNotSame(original, copy);
    }

    @Test
    public final void copyShouldBeEquivalentToOriginal() {
        assertEquals(original, copy);
    }

    @Test
    public final void copyShouldHaveObjectFieldsWhichAreNotSameAsOriginal() throws Exception {
        for (Class<?> clazz = original.getClass(); clazz != null; clazz = clazz.getSuperclass())
            compareFieldsFrom(clazz);
    }

    /**
     * @return an instance of a cloneable object
     */
    protected abstract T instance() throws Exception;

    /**
     * @param object a cloneable object -- this will be the result of having called {@link #instance()}.
     * @return an instance of a cloneable object; this should be the result of calling clone() on {@code object}.
     */
    protected abstract T copy(T object) throws Exception;

    /**
     * @return a collection of classes that the subclass wants us to treat as immutable.  Fields of these types
     * will not be compared for identity, but will be compared for equality.  Currently, the immutable types supported
     * out-of-the-box are {@link String}, the primitive wrapper classes, {@link BigInteger}, and {@link BigDecimal}.
     * Override this method to supply more classes.
     */
    protected Collection<Class<?>> classesToTreatAsImmutable() {
        return emptySet();
    }

    private boolean isImmutable(Class<?> fieldClass) {
        return fieldClass.isPrimitive() || immutableClasses.contains(fieldClass);
    }

    @SuppressWarnings("unchecked")
    private Set<Class<?>> coreImmutableClasses() {
        Set<Class<?>> immutables = new HashSet<Class<?>>();
        addAll(immutables, String.class, Boolean.class, Byte.class, Character.class, Double.class, Float.class,
            Integer.class, Long.class, Short.class, BigInteger.class, BigDecimal.class);
        return immutables;
    }

    private void compareFieldsFrom(Class<?> clazz) throws Exception {
        for (Field each : clazz.getDeclaredFields())
            compareField(each);
    }

    private void compareField(Field field) throws Exception {
        Class<?> fieldClass = field.getType();
        if (isImmutable(fieldClass))
            return;

        field.setAccessible(true);
        String fieldName = field.getDeclaringClass().getName() + '#' + field.getName();

        Object originalValue = field.get(original);
        Object copyValue = field.get(copy);
        assertNotSame(fieldName, originalValue, copyValue);
        if (fieldClass.isArray())
            compareElementsOfArray(fieldName, originalValue, copyValue);
        else if (List.class.isAssignableFrom(fieldClass))
            compareElementsOfList(fieldName, (List<?>) originalValue, (List<?>) copyValue);
        else if (Map.class.isAssignableFrom(fieldClass))
            compareValuesOfMap(fieldName, (Map<?, ?>) originalValue, (Map<?, ?>) copyValue);
    }

    private void compareElementsOfArray(String fieldName, Object originalArray, Object copyArray) {
        for (int i = 0, length = getLength(originalArray); i < length; ++i) {
            Object originalElement = get(originalArray, i);
            Object copyElement = get(copyArray, i);
            if (!isImmutable(originalElement.getClass()))
                assertNotSame(fieldName + " element " + i, originalElement, copyElement);
        }
    }

    private void compareElementsOfList(String fieldName, List<?> originalList, List<?> copyList) {
        for (int i = 0, length = originalList.size(); i < length; ++i) {
            Object originalElement = originalList.get(i);
            Object copyElement = copyList.get(i);
            if (!isImmutable(originalElement.getClass()))
                assertNotSame(fieldName + " element " + i, originalElement, copyElement);
        }
    }

    private void compareValuesOfMap(String fieldName, Map<?, ?> originalMap, Map<?, ?> copyMap) {
        for (Object key : originalMap.keySet()) {
            Object originalValue = originalMap.get(key);
            Object copyValue = copyMap.get(key);
            if (!isImmutable(originalValue.getClass()))
                assertNotSame(fieldName + " key " + key, originalValue, copyValue);
        }
    }
}
