/*
Continuous Testing Toolkit
Copyright (C) 2009 Paul R. Holser, Jr.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

package org.infinitest.toolkit;

import java.util.Collection;
import java.util.Map;
import java.util.Map.Entry;

import static com.google.common.collect.Maps.*;
import static java.util.Collections.*;

import org.infinitest.toolkit.runners.EnclosedInners;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import static org.junit.Assert.*;

/**
 * Support structure for testing the {@linkplain java.util.Collections#unmodifiableMap(Map) unmodifiability} of a map.
 * This is useful for verifying that accessor methods that return maps do not allow direct adding or removing from the
 * underlying map field.
 *
 * @param <K> the type of the keys in the maps involved
 * @param <V> the type of the values in the maps involved
 * @author <a href="http://www.pholser.com">Paul Holser</a>
 */
@RunWith(EnclosedInners.class)
public abstract class UnmodifiableMapTestSupport<K, V> {
    private Map<K, V> entries;
    private K containedKey;
    private V containedValue;

    @Before
    public final void initializeHarness() {
        this.entries = newMap();
        this.containedKey = containedKey();
        this.containedValue = containedValue();

        assertFalse("need a non-empty map", entries.isEmpty());
        assertTrue("need a key that is in the map", entries.containsKey(containedKey));
        assertTrue("need a value that is in the map", entries.containsValue(containedValue));
    }

    @Test(expected = UnsupportedOperationException.class)
    public final void shouldPreventPut() {
        entries.put(newKey(), null);
    }

    @Test(expected = UnsupportedOperationException.class)
    public final void shouldPreventPutAll() {
        Map<K, V> singletonMap = singletonMap(newKey(), null);
        entries.putAll(singletonMap);
    }

    @Test(expected = UnsupportedOperationException.class)
    public final void shouldPreventClear() {
        entries.clear();
    }

    @Test(expected = UnsupportedOperationException.class)
    public final void shouldPreventRemove() {
        entries.remove(containedKey);
    }

    /**
     * @return a map to test for unmodifiability
     */
    protected abstract Map<K, V> newMap();

    /**
     * @return a key to attempt to add to the {@linkplain #newMap() map}
     */
    protected abstract K newKey();

    /**
     * @return a key that is in the {@linkplain #newMap() map}
     */
    protected abstract K containedKey();

    /**
     * @return a value that is in the {@linkplain #newMap() map}
     */
    protected abstract V containedValue();

    class KeySetTest extends UnmodifiableCollectionTestSupport<K> {
        @Override
        protected K containedItem() {
            return containedKey;
        }

        @Override
        protected Collection<K> newCollection() {
            return entries.keySet();
        }

        @Override
        protected K newItem() {
            return newKey();
        }
    }

    class ValuesTest extends UnmodifiableCollectionTestSupport<V> {
        @Override
        protected V containedItem() {
            return containedValue;
        }

        @Override
        protected Collection<V> newCollection() {
            return entries.values();
        }

        @Override
        protected V newItem() {
            return null;
        }
    }

    class EntriesTest extends UnmodifiableCollectionTestSupport<Map.Entry<K, V>> {
        @Override
        protected Entry<K, V> containedItem() {
            return immutableEntry(containedKey, containedValue);
        }

        @Override
        protected Collection<Entry<K, V>> newCollection() {
            return entries.entrySet();
        }

        @Override
        protected Entry<K, V> newItem() {
            return null;
        }
    }
}
